from gimmick import constants
from gimmick.models import autoencoder_dense
import tensorflow as tf

algo_mapping = {
    'autoencoder_dense': autoencoder_dense.Model()
}

optimizer_mapping = {
    'adam': tf.keras.optimizers.Adam(learning_rate=constants.DEFAULT_LEARNING_RATE),
    'adamax': tf.keras.optimizers.Adamax(learning_rate=constants.DEFAULT_LEARNING_RATE),
    'nadam': tf.keras.optimizers.Nadam(learning_rate=constants.DEFAULT_LEARNING_RATE),
    'adagrad': tf.keras.optimizers.Adagrad(learning_rate=constants.DEFAULT_LEARNING_RATE),
    'rmsprop': tf.keras.optimizers.RMSprop(learning_rate=constants.DEFAULT_LEARNING_RATE),
    'sgd': tf.keras.optimizers.SGD(learning_rate=constants.DEFAULT_LEARNING_RATE)
}

metrics_mapping = {
    'mse': tf.keras.metrics.MeanSquaredError(),
    'rmse': tf.keras.metrics.RootMeanSquaredError(),
    'mae': tf.keras.metrics.MeanAbsoluteError(),
    'mape': tf.keras.metrics.MeanAbsolutePercentageError(),
    'msle': tf.keras.metrics.MeanSquaredLogarithmicError(),
    'mean': tf.keras.metrics.Mean(),
    'sum': tf.keras.metrics.Sum(),
    'kldiv': tf.keras.metrics.KLDivergence()
}

loss_function_mapping = {
    'mse': tf.keras.losses.MeanSquaredError(),
    'mae': tf.keras.losses.MeanAbsoluteError(),
    'mape': tf.keras.losses.MeanAbsolutePercentageError(),
    'msle': tf.keras.losses.MeanSquaredLogarithmicError(),
    'kldiv': tf.keras.losses.KLDivergence(),
    'cosine': tf.keras.losses.CosineSimilarity()
}


def convert_2dto3d(images):
    print("convering images to 3d: shape before convert_2dto3d", images.shape)
    if len(images.shape) == 3:
        images = images.reshape(images.shape[0], images.shape[1], images.shape[2], 1)
    print("shape after convert_2dto3d", images.shape)
    return images

if __name__ == "__main__":
    from sklearn import datasets
    digits = datasets.load_digits()
    images = digits.images  # It contains roughly 1800 images of shape 8 x 8
    print("shape before convert_2dto3d", images.shape)
    images = convert_2dto3d(images)
    print("shape after convert_2dto3d", images.shape)



import pickle
import tensorflow as tf
from gimmick.distributer import *
from gimmick import mapping

def load(modelfile):
    modelfile_tf = "tf_" + modelfile.split('.')[0] + ".h5"

    modelobj = pickle.load(open(modelfile, 'rb'))
    modelobj.model = tf.keras.models.load_model(modelfile_tf)
    modelobj.metrics = [mapping.metrics_mapping.get(x) for x in modelobj.metrics_keys]
    modelobj.optimizer = mapping.optimizer_mapping.get(modelobj.optimizer_keys)
    modelobj.loss_function = mapping.loss_function_mapping.get(modelobj.loss_function_keys)
    return modelobj


import math
import pickle
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.callbacks import ModelCheckpoint
from datetime import datetime 
from gimmick import constants


class Model():
    def __init__(self, learning_rate=None, optimizer=None, optimizer_keys=None, loss_function=None, loss_function_keys=None, metrics=None, metrics_keys=None,
                 code_length=8, num_encoder_layer='auto', num_decoder_layers='auto'):
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.loss_function = loss_function
        self.metrics = metrics
        self.num_encoder_layer = num_encoder_layer
        self.num_decoder_layers = num_decoder_layers
        self.code_length = code_length

        self.loss_function_keys = loss_function_keys
        self.optimizer_keys = optimizer_keys
        self.metrics_keys = metrics_keys

    ''' This function build model graph  '''
    def build_model_graph(self, images_shape):
        total_pixels = images_shape[0] * images_shape[1] * images_shape[2]
        
        num_encoder_layers = int(math.log(images_shape[0], 2)) - 2 if self.num_encoder_layer == "auto" else int(self.num_encoder_layer)
        num_encoder_layers = max(num_encoder_layers, 3)

        num_decoder_layers = int(math.log(images_shape[0], 2))- 2 if self.num_decoder_layers == "auto" else int(self.num_decoder_layers)
        num_decoder_layers = max(num_decoder_layers, 3)
        
        log2_code = int(math.log(self.code_length, 2))
        print("num_enoder_layer:\t", num_encoder_layers)
        print("num_decoder_layers:\t", num_decoder_layers)
        print("log2_code:\t\t", log2_code)

        model = keras.Sequential(name="model_dense")
        model.add(keras.Input(shape=images_shape, dtype=tf.int8))
        model.add(layers.Flatten())
        
        # Encoder Layer
        for i in range(1, num_encoder_layers + 1):
            neurons = 2 ** (num_encoder_layers - i + log2_code + 1) # Encoder layer size will be always greater then code_length by multiple of 2
            model.add(layers.Dense(neurons, activation="relu", name="encoder_layer_" + str(i) ))

        # Code Layer
        model.add(layers.Dense(self.code_length, name="code"))

        # Decoder Layer
        for i in range(1, num_decoder_layers + 1):
            neurons = 2 ** (i + log2_code)  # Decoder layer size will be always greater then code_length by multiple of 2
            model.add(layers.Dense(neurons, activation="relu", name="decoder_layer_" + str(i) ))

        model.add(layers.Dense(total_pixels, activation="relu", name="final_layer"))
        model.add(layers.Reshape(images_shape))
        
        optimizer =self.optimizer
        optimizer.learning_rate = self.learning_rate
        
        model.compile(optimizer=optimizer, loss=self.loss_function, metrics=self.metrics)
        
        print(model.summary())
        self.model = model

    ''' This function train model '''
    def train(self, images_train, images_test, epochs=10, batch_size=16, validation_split=0.2):

        startime = datetime.now()
        
        checkpoint = ModelCheckpoint(constants.DEFAULT_TF_MODELFILE, verbose=0, monitor='val_loss', save_best_only=True, mode='auto')
        early_stopping = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=5, restore_best_weights=True)

        print("================================= Training ===================================")
        model = self.model
        model.fit(images_train, images_train, batch_size=batch_size, epochs=epochs, validation_split=validation_split, 
                  callbacks=[checkpoint, early_stopping], shuffle=True)
        
        model.save(constants.DEFAULT_TF_MODELFILE) # Save Best model to disk
        print("Total Training time:", datetime.now() - startime)

        print("================================= Evaluating ===================================")
        model.evaluate(images_test, images_test, batch_size=batch_size, verbose=True)
        
    def prepare_code_statistics(self, images, batch_size=8, sample_size=64):
        ''' This function return the statistics for intermedite highly condence space of N Dimention 
        which can be used to generate similar samples '''
        print("================================= generating code statistics ===================================")

        print("Total samples used to generate code statistics:", sample_size)
        model = self.model
        images_shape = images[0].shape
        
        layers = [keras.Input(shape=images_shape, dtype=tf.int8)]
        encoder_layers = [layer.name if 'encoder_layer' in layer.name else None for layer in model.layers]
        num_encoder_layers = len(list(filter(lambda x: x, encoder_layers))) + 2 # 1 for Flatten layer and 1 for code layer
        layers.extend(model.layers[:num_encoder_layers])  # Trim all layers except encoder layers

        model_code_generator = keras.Sequential(layers)
        model_code_generator.build((None, images_shape[0], images_shape[1], images_shape[2]))

        for layer in model_code_generator.layers:
            if list(filter(lambda x: x in layer.name, ['flatten', 'reshape'])):
                continue
            assert all([np.array_equal(layer.get_weights()[0], model.get_layer(layer.name).get_weights()[0]), 
                        np.array_equal(layer.get_weights()[1], model.get_layer(layer.name).get_weights()[1])]),  "%s weights not same" % layer.name

        print(model_code_generator.summary())

        codes = model_code_generator.predict(images[:sample_size], batch_size=batch_size, verbose=False)
        print("codes shape:", codes.shape)

        assert codes.shape[1] == self.code_length, "code_length_passed (%d) and code_length_generated (%d) does not match" % (self.code_length, codes.shape[1])
        
        print(codes[0].tolist())
        print(codes[1].tolist())
        print(codes[2].tolist())
        
        code_stats = { 
            "min" : np.min(codes), 
            "max" : np.max(codes), 
            "mean": np.mean(codes),
            "std": np.std(codes)
        }
        self.code_stats = code_stats
        print("code_stats:", code_stats)

        
    ''' This function generate samples based on code statistics '''
    def generate(self, n, batch_size=8):
        print("================================= generating samples ===================================")
        model = self.model
        code_stats = self.code_stats
        
        encoder_layers = [layer.name if 'encoder_layer' in layer.name else None for layer in model.layers]
        num_encoder_layers = len(list(filter(lambda x: x, encoder_layers))) + 2 # 1 for Flatten layer and 1 for code layer
        
        # Building model
        model_generator = keras.Sequential(model.layers[num_encoder_layers:])
        model_generator.build((None, self.code_length))
        print(model_generator.summary())

        inputs  = np.random.normal(code_stats['mean'], code_stats['std'], (n, self.code_length))  # Random samples

        image_generated = model_generator.predict(inputs, batch_size=batch_size, verbose=False).astype(np.uint8)
        image_generated[image_generated > 255] = 255
        image_generated[image_generated < 0] = 0
        return image_generated
    
    def save(self, modelfile):
        modelfile_tf = "tf_" + modelfile.split('.')[0] + ".h5"
        self.model.save(modelfile_tf)
        
        model = self.model
        metrics = self.metrics

        self.model = None
        self.metrics = None
        self.optimizer = None
        self.loss_function = None
        
        print("Pickle protocol:", pickle.HIGHEST_PROTOCOL)
        with open(modelfile, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
        
        self.model = model
        self.metrics = metrics


import pickle
from gimmick import mapping
from gimmick.image_op import functions as image_functions
from sklearn.model_selection import train_test_split

''' 
This function create and fit model based on passed arguments 
Paramters:

images :                     Images to be passed to learning functions, has shape [N, (2D or 3D)], 
                             where N is number of samples and 2D and 3D denotes image size
algo :                       Algorithm to be used to learn image representation, Eg Autoencode_dense, 
code_length:                 Default 8, Length of intermediate representation or condense space generated 
                             by model. In order to generate a random image sample having dimention equal 
                             to code_length must be passed.
epochs:                      Default 10, number of epoch to be used while training model
batch_size:                  Default 16, batch size of each training/eval/generation step
optimizer:                   Default 'adam', optimizer used to train the model
learning_rate:               Default 0.001, learning rate for training model
loss_function:               Default 'mse', loss function for training model
metrics:                     Default ['mse'], list of metrics to be printed while training
num_encoder_layer:           Default 'auto', number of layers to be used in encoder
num_decoder_layers:          Default 'auto', number of layers to be used in decoder
samples_for_code_statistics: Default 64, samples to be used to generate code statistics
'''
def learn(images, algo, code_length=8, num_encoder_layer='auto', num_decoder_layers='auto', epochs=10, 
          batch_size=16, optimizer='adam', learning_rate=0.001, loss_function='mse', metrics=['mse'],
          samples_for_code_statistics=64):

    print("Number sample:", images.shape[0], "Image shape:", images[0].shape)

    model = mapping.algo_mapping.get(algo, None)
    if not model:
        raise Exception("Algo not implement/present possible values for also are %s" % (mapping.algo_mapping.keys()))

    optimizer_keys = optimizer
    optimizer = mapping.optimizer_mapping.get(optimizer, None)
    if not optimizer:
        raise Exception("Optimizer not implement/present possible values for also are %s" % (mapping.optimizer_mapping.keys()))

    loss_function_keys = loss_function
    loss_function = mapping.loss_function_mapping.get(loss_function, None)
    if not optimizer:
        raise Exception("loss_function not implement/present possible values for also are %s" % (mapping.loss_function_mapping.keys()))

    metrics_keys = metrics
    metrics = [mapping.metrics_mapping.get(x, None) for x in metrics]
    if list(filter(lambda x: x == None, metrics)):
        raise Exception("One or more of the metrics passed is not a valid metrics, possible values are %s" % (mapping.metrics_mapping.keys()))
        
    print("===================================================================")
    print("Algo:\t\t", model)
    print("optimizer:\t", optimizer)
    print("loss_function:\t", loss_function)
    print("metrics:\t", metrics)
    print("Epochs:\t\t", epochs)
    print("batch_size:\t", batch_size)
    print("learning_rate:\t", learning_rate)

    print("===================================================================")

    ## Write code to normalize image to nearest 2 power
    images = image_functions.convert_2dto3d(images)

    model.__init__(learning_rate=learning_rate, optimizer=optimizer, optimizer_keys=optimizer_keys, 
                   loss_function=loss_function, loss_function_keys=loss_function_keys, 
                   metrics=metrics, metrics_keys=metrics_keys,code_length=code_length, 
                   num_encoder_layer=num_encoder_layer, num_decoder_layers=num_decoder_layers)
    model.build_model_graph(images[0].shape)
    
    images_train, images_test = train_test_split(images, test_size=0.3, shuffle=True)

    print("Train:", len(images_train.shape))
    print("Test:", len(images_test.shape))

    model.train(images_train, images_test, epochs=epochs, batch_size=batch_size, validation_split=0.2)
    model.prepare_code_statistics(images_train, batch_size=batch_size, sample_size=samples_for_code_statistics)
    return model


import setuptools
from glob import glob

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gimmick",
    version="0.2",
    author="Pankaj Rawat",
    author_email="pankajr141@gmail.com",
    description="Libraray contains algo to generate images by learning representation from data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pankajr141/gimmick",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
        'tensorflow>=2.3.1',
        'matplotlib',
        'scikit-learn==0.24.2'
    ],
)
